(in-package :cube-modeller)

(defvar *animation-pathname* nil)

(defmacro make-button (name label function)
  `(make-pane 'push-button-pane
              :name ',(read-from-string (format nil "~s-button" name))
              :label ,label
              :activate-callback (lambda (&rest rest)
                                   (declare (ignore rest))
                                   ,function)))

(defmacro make-move-button (name &optional key)
  `(make-button ,name
                ;; ,(string-downcase (symbol-name name))
                ,(format nil "~a~@[ (~(~s~))~]" (command-name-from-symbol name) key)
                (,(read-from-string (format nil "move-~s" name)))))

(defmacro make-color-button (color)
  `(make-button ,(read-from-string (format nil "set-intensity-to-~s" color))
                ,(format nil "~s" color)
                (set-current-intensity ,color)))

(defmacro make-goto-cube-button (name &optional key)
  `(make-button ,name
                ;; ,(string-downcase (symbol-name name))
                ,(format nil "~a~@[ (~(~s~))~]" (command-name-from-symbol name) key)
                (goto-cube ,(read-from-string (format nil ":~s" name)))))

(defmacro make-command-button (name command &optional key)
  `(make-button ,name
                ,(format nil "~a~@[ (~(~s~))~]" (command-name-from-symbol name) key)
                (throw-command ,command)))


;;; APPLICATION-FRAME

(define-application-frame toolbox ()
  ()
  (:panes
   (cube-slider :slider
                :id 'cube-slider-gadget
                :min-value 0
                :max-value 0
                :value 0
                ;; :show-value-p t
                :orientation :horizontal
                :current-color +black+)
   (cube-number-pane :application-pane ;; this was just a workaround to make the menus work
                     :min-height 15
                     :background climi::*3d-normal-color*
                     :display-function
                     (lambda (frame pane)
                       (declare (ignore frame))
                       (write-string-centered (format nil "Cube ~d of ~d" *current-cube-number*
                                                                          (1- (length *animation*)))
                                              pane)))
   (transformations-pane (make-pane 'option-pane
                                    :width 200
                                    :items (reverse *transformations*)
                                    :name-key #'command-name-from-symbol
                                    :value (car (reverse *transformations*)))))
   (:layouts
   (defaults (vertically ()
               (labelling (:label "Files")
                 (horizontally ()
                   +fill+
                   (make-command-button load-animation 'com-load-animation-file #+nil l)
                   (make-command-button save-animation 'com-save-animation-file #+nil s)
                   +fill+))
                                                                                            
               (labelling (:label "Playing")
                 (horizontally ()
                   +fill+
                   (make-button toggle-play-animation "Toggle play (p)" (toggle-animation-running-state))
                   +fill+))
               
               (labelling (:label "Goto cube")
                 (vertically ()
                   (horizontally ()
                     +fill+
                     (vertically ()
                       (make-goto-cube-button first home)
                       (make-goto-cube-button last end))
                     +fill+
                     (vertically ()
                       (make-goto-cube-button previous pg-up)
                       (make-goto-cube-button next pg-down))
                     +fill+)
                   15
                   cube-number-pane
                   cube-slider))

               (labelling (:label "General")
                 (horizontally ()
                   +fill+
                   (make-button undo "undo (u)" (undo))
                   +fill+
                   (make-button reset-perspective "Reset perspective (r)" (reset-perspective))
                   +fill+))

               (labelling (:label "Shade of red")
                 (horizontally ()
                   +fill+
                   (make-color-button 0)
                   (make-color-button 1)
                   (make-color-button 2)
                   (make-color-button 3)
                   (make-color-button 4)
                   (make-color-button 5)
                   +fill+))

               (labelling (:label "Change all LEDs of this cube")
                 (horizontally ()
                     +fill+
                     (make-button delete-cube "Set to current color (f)" (fill-cube))
                     (make-button delete-cube "Switch them off (e)" (empty-cube))
                     +fill+))

               (labelling (:label "Insert cube")
                 (horizontally ()
                   +fill+
                   (make-button insert-before "Before current (b)" (insert-cube-copy :position :before))
                   (make-button insert-after "After current (a)" (insert-cube-copy))
                   +fill+))

               (labelling (:label "Delete...")
                 (horizontally ()
                   +fill+
                   (make-button delete-cube "this cube (d)" (delete-cube))
                   (make-button delete-animation "the whole animation (k)" (delete-animation))
                   +fill+))
               
               (labelling (:label "Move the current cube")
                 (horizontally ()
                   +fill+
                   (vertically ()
                     +fill+
                     (make-move-button up)
                     (make-move-button down)
                     +fill+)
                   +fill+
                   (vertically ()
                     (make-move-button forward s)
                     (horizontally ()
                       (50 (make-move-button left))
                       (50 (make-move-button right)))
                     (make-move-button backward x))
                   +fill+))
               
               (labelling (:label "Transformations")
                 (horizontally ()
                   +fill+
                   transformations-pane
                   (make-button 'transformate "Transform"
                                (progn
                                  (funcall (gadget-value (find-pane-named *application-frame*
                                                                          'transformations-pane)))
                                  (write-cube-to-device)
                                  (adjust-cube-slider-for-animation)))
                   +fill+))
       
               ))))

(defmethod run-frame-top-level :before ((frame toolbox) &key)
  (setf *toolbox-frame* frame)
  (adjust-cube-slider-for-animation))

(defmethod run-frame-top-level :after ((frame toolbox) &key)
  (declare (ignore frame))
  (setf *toolbox-frame* nil))

(climi::make-run-gui-defun toolbox
                           :function-name start-toolbox
                           :default-process-name "cube-modeller toolbox")


;;; COMMANDS

(define-toolbox-command (com-exit :name "Quit" #+nil "Quit (q)" :menu t
                          :keystroke (#\q))
    ()
  (frame-exit *application-frame*))

(define-toolbox-command (com-load-animation-file :name t
                                                 :keystroke (#\l))
    ()
  (load-animation-file
   (setf *animation-pathname*
         (file-selector:select-file :own-window t
                                    :pathname (or *animation-pathname*
                                                  *default-pathname*))))
  (adjust-cube-slider-for-animation))

(define-toolbox-command (com-save-animation-file :name t
                                                 :keystroke (#\s))
    ()
  (save-animation-file
   (setf *animation-pathname*
         (file-selector:select-file :own-window t
                                    :pathname (or *animation-pathname*
                                                  *default-pathname*)))
   :if-exists :overwrite)
  (adjust-cube-slider-for-animation))


;;; CALLBACKS

(defmethod value-changed-callback ((gadget slider-pane) (client toolbox) (gadget-id (eql 'cube-slider-gadget)) value)
  (declare (ignore gadget client gadget-id))
  (goto-cube (round value))
  ;; NEEDS-REDISPLAY was not enough
  ;; (setf (pane-needs-redisplay (find-pane-named *application-frame* 'cube-number-pane)) t)
  (redisplay-frame-pane *application-frame* (find-pane-named *application-frame* 'cube-number-pane)))

(defmethod drag-callback ((pane slider-pane) (client toolbox) (gadget-id (eql 'cube-slider-gadget)) value)
  (value-changed-callback pane client gadget-id value))


;;; setters for the cube-slider

(defun set-cube-slider-value (&optional (value *current-cube-number*))
  (setf (gadget-value (find-pane-named *application-frame* 'cube-slider)) ;; :invoke-callback t)
        value)
  ;; The redisplay code has to be copied here; we cannot use :invoke-callback t anymore, as
  ;; then there is a infinitive loop (GOTO-CUBE calls SET-CUBE-SLIDER, this triggers the
  ;; VALUE-CHANGED-CALLBACK that invokes GOTO-CUBE...).
  
  ;; NEEDS-REDISPLAY was not enough
  ;; (setf (pane-needs-redisplay (find-pane-named *application-frame* 'cube-number-pane)) t)
  (redisplay-frame-pane *application-frame* (find-pane-named *application-frame* 'cube-number-pane)))

(defun adjust-cube-slider-for-animation (&key (animation *animation*))
  (let ((value (length animation)))
    (setf (gadget-max-value (find-pane-named *application-frame* 'cube-slider))
          (1- value)
          
          (slot-value (find-pane-named *application-frame* 'cube-slider) 'climi::number-of-quanta)
          value))
  (set-cube-slider-value *current-cube-number*))
